/*
    This program scans the user part of virtual memory space and
    classifies each contigous block of pages that have the same status.

    This program takes a PID on the command line (and optional upper and
    lower bounds on the address space) and scans the memory space of the
    process with that PID.
*/
#include <windows.h>
#include <stdio.h>
#include <inttypes.h>
// set the beginning and ending addresses of the memory map
// these can be used to save space in the console window
// these addresses should be on 64K allocation region boundries
// the maximum value for DEFAULT_END_MEMORY_MAP_ADDRESS is 0x7fff0000000 (43 bits)
#define DEFAULT_BEGIN_MEMORY_MAP_ADDRESS 0x00000000000ULL
#define DEFAULT_END_MEMORY_MAP_ADDRESS   0x7fff0000000ULL

// function prototypes
void memoryMapper(HANDLE, void*, void*);
void printRegionInfo(MEMORY_BASIC_INFORMATION);
void accumulateStatistics(MEMORY_BASIC_INFORMATION, ULONGLONG*, ULONGLONG*, ULONGLONG*);
void printStatistics(ULONGLONG, ULONGLONG, ULONGLONG);
void printError(char*);


/*************************************************************************************
    The main method gets a handle for the process to map by using the processes PID
    from the command line. Then main calls the memoryMapper function, printing
    the results to standard output.
*/
int main(int argc, char *argv[])
{
   HANDLE processHandle; // handle for process to be mapped
   int processID;        // pid of process to be mapped

   void* begin_memory_map_address = (void*)DEFAULT_BEGIN_MEMORY_MAP_ADDRESS;
   void*   end_memory_map_address = (void*)DEFAULT_END_MEMORY_MAP_ADDRESS;

   if (argc < 2)
   {
      fprintf(stderr, "Usage: %s <pid> [<begin address>] [<end address>]", argv[0]);
      return -1;
   }
   else
   {
      processID = atoi(argv[1]);
      processHandle = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, processID);
      if (processHandle == NULL)
      {
          printError("OpenProcess");
          return -1;
      }
   }

   if (argc > 2)  // the beginning address should be rounded to an allocation boundary
      begin_memory_map_address = (void*)strtoull(argv[2], NULL, 16);
   if (argc > 3)
        end_memory_map_address = (void*)strtoull(argv[3], NULL, 16);

   // map the process's virtual memory space
   memoryMapper(processHandle, begin_memory_map_address, end_memory_map_address);

   return 0;
}//main()


/********************************************************************************
     This function walks its way through virtual memory getting information
     about each distinct region of memory by using the function VirtualQueryEx.

     This function presupposes that the parameter value for begin_memory_map_address
     is at a 64KB allocation region boundary.
*/
void memoryMapper(HANDLE processHandle,
                  void* begin_memory_map_address,
                  void* end_memory_map_address)
{
   MEMORY_BASIC_INFORMATION mbi;
   ULONGLONG reserved = 0, committed = 0, free = 0;
   void* regionAddress = begin_memory_map_address;

   while (regionAddress < end_memory_map_address)
   {
      // get information about the current region in virtual memory
      if (VirtualQueryEx(processHandle, regionAddress, &mbi, sizeof(mbi)) != 0)
      {
         printRegionInfo(mbi);
         accumulateStatistics(mbi, &reserved, &committed, &free);
         // point regionAddress at the beginning of the next region in virtual memory
         regionAddress += mbi.RegionSize;
      }
      else
      {
         printError("VirtualQueryEx");
         ExitProcess(-1);
      }
   }
   printStatistics(reserved, committed, free);
}


/****************************************************************************
    This function will be called once for every distinct block of virtual memory.
    This function formats the information from a MEMORY_BASIC_INFORMATION data
    structure so that the information can be printed out. This function will
    print one line of information to standard output. If the information is
    from a new allocation region or from a free region, then a blank line is
    also printed to the output.
*/
void printRegionInfo(MEMORY_BASIC_INFORMATION mbi)
{
   static void* oldAllocationBase; // used to determine allocation region boundries
   char* state = NULL;
   char* type = NULL;
   char* protect = NULL;

   if ( (mbi.State != MEM_FREE) && (mbi.AllocationBase != oldAllocationBase) )
   {
      printf("\n");  // blank line separates allocation regions
   }

   if (mbi.State != MEM_FREE)  // find either a reserved or a committed block of pages
   {
      oldAllocationBase = mbi.AllocationBase;  //remember the current allocation base

      // determine the type of the block
      if      (mbi.Type == MEM_PRIVATE) type = "private";
      else if (mbi.Type == MEM_MAPPED)  type = "mapped";
      else if (mbi.Type == MEM_IMAGE)   type = "image";

      // determine the state of the block of pages
      if      (mbi.State == MEM_COMMIT)  state = "commit";
      else if (mbi.State == MEM_RESERVE) state = "reserve";

      // determine the protection status of the block
      protect = "______\t";  // a default value
      if (mbi.State == MEM_COMMIT)
      {
         if      (mbi.Protect == PAGE_NOACCESS)          protect = "no-access";
         else if (mbi.Protect == PAGE_READONLY)          protect = "read-only";
         else if (mbi.Protect == PAGE_READWRITE)         protect = "read-write";
         else if (mbi.Protect == PAGE_WRITECOPY)         protect = "write-copy";
         else if (mbi.Protect == PAGE_EXECUTE)           protect = "execute";
         else if (mbi.Protect == PAGE_EXECUTE_READ)      protect = "execute-read";
         else if (mbi.Protect == PAGE_EXECUTE_READWRITE) protect = "execute-read-write";
         else if (mbi.Protect == PAGE_EXECUTE_WRITECOPY) protect = "execute-write-copy";
         else if (mbi.Protect == (PAGE_READONLY|PAGE_GUARD))
                                                         protect = "read-only-pageGuard";
         else if (mbi.Protect == (PAGE_READWRITE|PAGE_GUARD))
                                                         protect = "read-write-pageGuard";
         else if (mbi.Protect == (PAGE_WRITECOPY|PAGE_GUARD))
                                                         protect = "write-copy-pageGuard";
         else if (mbi.Protect == (PAGE_EXECUTE|PAGE_GUARD))
                                                         protect = "execute-pageGuard";
         else if (mbi.Protect == (PAGE_EXECUTE_READ|PAGE_GUARD))
                                                         protect = "execute-read-pageGuard";
         else if (mbi.Protect == (PAGE_EXECUTE_READWRITE|PAGE_GUARD))
                                                         protect = "execute-read-write-pageGuard";
         else if (mbi.Protect == (PAGE_EXECUTE_WRITECOPY|PAGE_GUARD))
                                                         protect = "execute-write-copy-pageGuard";
      }
      else if (mbi.State == MEM_RESERVE)
      {
         if      (mbi.AllocationProtect == PAGE_NOACCESS)     protect = "no-access";
         else if (mbi.AllocationProtect == PAGE_READONLY)     protect = "read-only";
         else if (mbi.AllocationProtect == PAGE_READWRITE)    protect = "read-write";
         else if (mbi.AllocationProtect == PAGE_WRITECOPY)    protect = "write-copy";
         else if (mbi.AllocationProtect == PAGE_EXECUTE)      protect = "execute";
         else if (mbi.AllocationProtect == PAGE_EXECUTE_READ) protect = "execute-read";
         else if (mbi.AllocationProtect == PAGE_EXECUTE_READWRITE)
                                                              protect = "execute-read-write";
         else if (mbi.AllocationProtect == PAGE_EXECUTE_WRITECOPY)
                                                              protect = "execute-write-copy";
      }

      // print out the info for the reserved or commited block
//    printf(           "%p size=%llu\t%s\t%s\t%s\n",
      printf("%#.11"PRIx64" size=%llu\t%s\t%s\t%s\n",
             (long long unsigned int)mbi.BaseAddress,
             (ULONGLONG)mbi.RegionSize,
             state, protect, type);
   }
   else  // found a free region
   {
      int flag;
      printf("\n"); // blank line separates free region
      if ( (ULONGLONG)mbi.RegionSize > 1000000000 ) // greater than 1 GB
    //if ( (ULONGLONG)mbi.RegionSize >  500000000 ) // greater than 0.5 GB
         flag = 1;
      else
         flag = 0;
      // print out the info for the free region
//    printf(           "%p size=%llu\t%s\n",
      printf("%#.11"PRIx64" size=%llu\t%s\n",
             (long long unsigned int)mbi.BaseAddress,
             (ULONGLONG)mbi.RegionSize,
             flag ? "****FREE****" : "free");
   }
}


/****************************************************************************
    This function is called once for every distinct region in the
    memory map. This function uses the data in the MEMORY_BASIC_INFORMATION
    block from the most recently mapped region to updates running totals for
    the amounts of reserved (but not committed), committed, and free memory.
*/
void accumulateStatistics(MEMORY_BASIC_INFORMATION mbi,
                          ULONGLONG* reserved, ULONGLONG* committed, ULONGLONG* free)
{
   if (mbi.State == MEM_RESERVE) *reserved += mbi.RegionSize;
   if (mbi.State == MEM_COMMIT) *committed += mbi.RegionSize;
   if (mbi.State == MEM_FREE)        *free += mbi.RegionSize;

}


/*******************************************************************************
    This function is called when a complete pass of the memory map is done.
    This function prints out the running totals kept by the accumulateStatistics
    function.
*/
void printStatistics(ULONGLONG reserved, ULONGLONG committed, ULONGLONG free)
{
   printf("\n");
   printf("Total  reserved: %15llu\n", reserved);
   printf("Total committed: %15llu\n", committed);
   printf("Total      free: %15llu\n", free);
   printf("Total    memory: %15llu\n", reserved+committed+free);
}


/*******************************************************************
   This function prints out "meaningful" error messages. If you call
   a Windows function and it returns with an error condition, then
   call this function right away and pass it a string containing the
   name of the Windows function that failed. This function will print
   out a reasonable text message explaining the error.
*/
void printError(char* functionName)
{
   LPSTR lpMsgBuf = NULL;
   int error_no;
   error_no = GetLastError();
   FormatMessage(
      FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      error_no,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* Default language */
      (LPTSTR)&lpMsgBuf,
      0,
      NULL
   );
   /* Display the string. */
   fprintf(stderr, "\n%s failed on error %d: %s", functionName, error_no, lpMsgBuf);
   MessageBox(NULL, lpMsgBuf, functionName, MB_OK);
   /* Free the buffer. */
   LocalFree( lpMsgBuf );
}
